#extension GL_OES_standard_derivatives : enable

varying highp vec4 var_position;
varying highp vec3 var_normal;
varying highp vec2 var_texcoord0;
varying highp vec4 var_texcoord0_shadow;

uniform highp sampler2D texture_diffuse;
uniform highp sampler2D texture_skybox;
uniform highp sampler2D texture_shadow;

// константы для пиксельного шейдера
uniform highp vec4 fog_data;
uniform highp vec4 camera_position;
uniform highp vec4 light;

const float SUN_POWER = 0.5;

#define PI 3.14159265

// Оператор тон-маппинга для преобразования HDR к LDR
highp vec4 ToneMapping(highp vec4 Color, highp float Exposure)
{
    return vec4(1.0 - exp(-Exposure*Color.rgb), Color.a);
}

// Получаем цвет и яркость освещения из сферической карты окружения (в проекции Меркатора)
// в направлении вектора Direction, который считается находящимся в мировом пространстве
highp vec4 Light_GetFromMerkatorSphericalMap(highp sampler2D Map, highp vec3 Direction, float bias)
{
    // получаем координату по y
    highp float vCos = dot(vec3(0.0,1.0,0.0), Direction);
    highp float vAngle = acos(vCos)/(PI*2.0);
    // получаем координату по x
    Direction.y = 0.0;
    Direction = normalize(Direction);
    highp float uCos = dot(vec3(1.0,0.0,0.0), Direction);
    highp float uAngle = acos(uCos)/(PI*2.0);
    if (Direction.z<0.0) uAngle = 1.0-uAngle;
    // получаем цвет из текстуры
    return textureLod(Map, vec2(uAngle, 1.0-4.0*vAngle), bias) * (1.0-vAngle*2.0);
}

// Получаем цвет тумана из сферической карты окружения (в проекции Меркатора)
// в направлении вектора Direction, который считается находящимся в мировом пространстве
highp vec4 Fog_GetFromMerkatorSphericalMap(highp sampler2D Map, highp vec3 Direction, float bias)
{
    // получаем координату по x
    Direction.y = 0.0;
    Direction = normalize(Direction);
    highp float uCos = dot(vec3(1.0,0.0,0.0), Direction);
    highp float uAngle = acos(uCos)/(PI*2.0);
    if (Direction.z<0.0) uAngle = 1.0-uAngle;
    // получаем цвет из текстуры
    return textureLod(Map, vec2(uAngle, 0.001), bias);
}

highp vec2 rand(vec2 co)
{
    return vec2(fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453), fract(sin(dot(co.yx ,vec2(12.9898,78.233))) * 43758.5453)) * 0.00047;
}

float shadow_calculation(highp vec4 depth_data)
{
    const highp float depth_bias = 0.0001;
    highp float shadow = 0.0;
    highp float texel_size = 1.0 / 4096.0;

    highp vec2 uv = depth_data.xy;
    highp vec4 depth = texture2D(texture_shadow, uv);
    shadow = (depth_data.z-depth_bias) > depth.x ? 1.0 : 0.0;

    if (uv.x<0.0) shadow = 0.0;
    if (uv.x>1.0) shadow = 0.0;
    if (uv.y<0.0) shadow = 0.0;
    if (uv.y>1.0) shadow = 0.0;
    
    return shadow;
}

highp vec4 float_to_rgba(highp float v)
{
    highp vec4 enc = vec4(1.0, 255.0, 65025.0, 16581375.0) * v;
    enc = fract(enc);
    enc -= enc.yzww * vec4(1.0/255.0, 1.0/255.0, 1.0/255.0, 0.0);
    return enc;
}

void main()
{
    highp vec3 sun_vec = vec3(0.015, 0.650, -0.760);
    highp vec3 view_vec = var_position.xyz - camera_position.xyz;
    highp float distance = length(view_vec);
    highp vec3 view_normalized = view_vec/distance;
    
    highp vec4 color = texture2D(texture_diffuse, var_texcoord0);
    
    highp vec3 real_normal = var_normal;
    float RNdotV = dot(real_normal, -view_normalized);
    if (RNdotV<0.0)
    {
        highp vec3 cross_normal = cross(real_normal, -view_normalized);
        real_normal = normalize( cross(cross_normal, -view_normalized) );
        if (dot(var_normal, real_normal)<0.0) real_normal = -real_normal;
    }

    vec4 ambient = Light_GetFromMerkatorSphericalMap(texture_skybox, real_normal, 9.0);
    vec4 fog = Fog_GetFromMerkatorSphericalMap(texture_skybox, view_normalized, 1.0);

    float TransmittedVolume = 0.94;
    highp float N = max((distance-8.0), 0.0);
    float fog_factor = pow(TransmittedVolume, N);

    highp vec4 shadow_projected = var_texcoord0_shadow / var_texcoord0_shadow.w;
    float shadow = 1.0 - shadow_calculation(shadow_projected);
    float sunlight = max(dot(real_normal, sun_vec), 0.0)*SUN_POWER*shadow;
    
    vec4 result_color = vec4(mix(fog, color*ambient + color*sunlight, fog_factor).rgb, color.a);
    result_color = ToneMapping(result_color, 2.0);
    
    // gl_FragData[0] = vec4(color.rgb, 1.0);
    gl_FragData[0] = vec4(result_color.rgb, 1.0);
    gl_FragData[1] = vec4(((real_normal/2.0)+0.5), 1.0);
    gl_FragData[2] = float_to_rgba(distance);
}
