local M = {}

-- настройки камеры
M.FOV = 60 -- 60 degrees (pi/3) FOV
M.z_min = 0.1
M.z_max = 1000
M.position = v3(0, 2, -4)
M.target = v3(0)
M.direction = -M.position
M.up_vector = v3(0, 1, 0)
M.axis_left = v3(-1, 0, 0)
M.rotation_X = 0
M.rotation_Y = 0
M.invert = 1 -- для инвертирования камеры, может принимать значения 1 или -1
M.sensivity = 200 -- для управления чувствительностью камеры
M.speed = 4 -- метра в секунду - скорость движения камеры

local current_time = 0
local need_update_camera = false
local TempRotX = 0
local TempRotY = 0

function M.update_rotations(rotation_X, rotation_Y)
	rotation_X = rotation_X or M.rotation_X
	rotation_Y = rotation_Y or M.rotation_Y
	-- поворачиваем вектор направления взгляда камеры (M.direction)
	local mRotY = vmath.matrix4_rotation_y(rotation_X)
	local direction = mRotY * v4(0, 0, 1, 0)
	local axis_left = mRotY * v4(-1, 0, 0, 0)
	M.axis_left = v3(axis_left.x, axis_left.y, axis_left.z)
	local mRotX = vmath.matrix4_axis_angle(M.axis_left, rotation_Y)
	direction = mRotX * direction
	M.direction = v3(direction.x, direction.y, direction.z)
end

function M.update_rotations_around_target(rotation_X, rotation_Y)
	rotation_X = rotation_X or M.rotation_X
	rotation_Y = rotation_Y or M.rotation_Y
	-- поворачиваем вектор направления взгляда камеры (M.direction)
	local mRotY = vmath.matrix4_rotation_y(rotation_X)
	local direction = mRotY * v4(0, 0, 1, 0)
	local axis_left = mRotY * v4(-1, 0, 0, 0)
	M.axis_left = v3(axis_left.x, axis_left.y, axis_left.z)
	local mRotX = vmath.matrix4_axis_angle(M.axis_left, rotation_Y)
	direction = mRotX * direction
	M.direction = v3(direction.x, direction.y, direction.z)

	local distance = vmath.length(M.position - M.target)
	M.direction = vmath.normalize(M.direction)
	M.position = M.target - (M.direction*distance)
end

local function rotation_clamp_bounds(rotation)
	-- проверяем на выход за определенные границы разумного (что б голову не запрокидывать назад)
	-- конкретно эти ограничения - для вращения камеры вокруг целевой точки
	if rotation<-1.5 then rotation=-1.5 end
	if rotation>-0.1 then rotation=-0.1 end
	return rotation
end

local function position_shift(direction)
	local speed = 0
	if M.last_time then speed = (current_time - M.last_time)*M.speed end
	M.position.x = M.position.x + speed*direction.x
	M.position.z = M.position.z + speed*direction.z
	M.last_time = current_time
	need_update_camera = true
end

local function target_shift(direction)
	local speed = 0
	if M.last_time then speed = (current_time - M.last_time)*M.speed end
	M.position.x = M.position.x + speed*direction.x
	M.position.z = M.position.z + speed*direction.z
	M.target.x = M.target.x + speed*direction.x
	M.target.z = M.target.z + speed*direction.z
	M.last_time = current_time
	need_update_camera = true
end

-- пересчитываем матрицы камеры из изменившихся исходных данных
function M.update()
	M.matrix_view = vmath.matrix4_look_at(M.position, M.position + M.direction, M.up_vector)
	M.matrix_proj = vmath.matrix4_perspective(3.141592/(180/M.FOV), SCREEN_WIDTH/SCREEN_HEIGHT, M.z_min, M.z_max)
end

local function normalize_2d(vector)
	local result = vector * 1.0
	if result.y then result.y = 0 end
	return vmath.normalize(result)
end

local function process_touch(action, number)
	if not M.touch then M.touch = {} end
	if not M.touch[number] then M.touch[number] = {} end
	-- сохраняем текущие координаты нажатия (они нужны только для визуализации в интерфейсе!)
	M.touch[number].a_Y = action.y
	M.touch[number].a_X = action.x
	-- нажали на экран
	if action.pressed then
		M.touch[number].X = action.x
		M.touch[number].Y = action.y
		M.touch[number].pressed = true
		-- определяем что именно мы нажали - область для перемещения или область для вращения камеры?
		if (action.x<WINDOW_WIDTH/2)and(DEVICE_INFO.system_name~="Windows") then
			M.touch[number].moving = true
		else
			M.touch[number].moving = false
		end
	end
	-- экран в состоянии нажатия
	if (M.touch[number].pressed)and(not action.released)and((M.touch[number].X~=action.x)or(M.touch[number].Y~=action.y)) then
		-- осуществляем сдвиг или вращение для камеры
		if M.touch[number].moving then
			local wf = (action.y-M.touch[number].Y)
			local ws = (action.x-M.touch[number].X)
			local v_forward = normalize_2d(M.direction) * wf / (math.abs(wf)+math.abs(ws))
			local v_side = normalize_2d(M.axis_left) * ws / (math.abs(wf)+math.abs(ws))
			target_shift(v_forward + v_side)
		else
			-- получаем временные углы поворота камеры
			TempRotX = M.rotation_X - ((action.x-M.touch[number].X)/M.sensivity)*M.invert
			TempRotY = M.rotation_Y + ((action.y-M.touch[number].Y)/M.sensivity)*M.invert
			TempRotY = rotation_clamp_bounds(TempRotY)
			M.update_rotations_around_target(TempRotX, TempRotY)
		end
		need_update_camera = true
	end
	-- отпустили нажатый экран
	if action.released then
		if M.touch[number].moving then
			M.touch[number].moving = false
		else
			M.rotation_X = TempRotX
			M.rotation_Y = TempRotY
			M.rotation_Y = rotation_clamp_bounds(M.rotation_Y)
			M.update_rotations_around_target()
		end
		need_update_camera = true
		M.touch[number] = nil
	end
end

function M.on_input(action_id, action)
	current_time = socket.gettime()
	-- обрабатываем мультитач
	if action_id==hash("multitouch") then
		for i, v in ipairs(action.touch) do
			process_touch(v, v.id)
		end
	elseif (action_id==hash("touch")) and (DEVICE_INFO.system_name=="Windows") then
		process_touch(action, 0)
	end
	-- отслеживаем движение камеры
	if action.pressed then
		if action_id==hash("key_w") then M.pressed_key_forward = true end
		if action_id==hash("key_s") then M.pressed_key_backward = true end
		if action_id==hash("key_a") then M.pressed_key_left = true end
		if action_id==hash("key_d") then M.pressed_key_right = true end
	end
	if action.released then
		if action_id==hash("key_w") then M.pressed_key_forward = false end
		if action_id==hash("key_s") then M.pressed_key_backward = false end
		if action_id==hash("key_a") then M.pressed_key_left = false end
		if action_id==hash("key_d") then M.pressed_key_right = false end
		-- сбрасываем время последней обработки, чтобы при старте нового движения не было рывков
		M.last_time = nil
	end

	local moving_direction
	if M.pressed_key_forward then moving_direction = (moving_direction or v3(0)) + normalize_2d(M.direction) end
	if M.pressed_key_backward then moving_direction = (moving_direction or v3(0)) - normalize_2d(M.direction) end
	if M.pressed_key_left then moving_direction = (moving_direction or v3(0)) - M.axis_left end
	if M.pressed_key_right then moving_direction = (moving_direction or v3(0)) + M.axis_left end
	if moving_direction then target_shift(moving_direction) end
	-- пересчитываем матрицы камеры из изменившихся исходных данных
	if need_update_camera then
		M.update()
		need_update_camera = false
	end
end

return M
